--
-- FS19 - Work RPM
-- This mod increases vehicle motor rpm when using hydraulics tools (moving tools), tipping, unfolding, lifting or rotating implement or moving pipe. Motor rpm is displayed on screen.
-- 				- you can resize or move rpm HUD
-- 				- you can open settings menu where you can set some options:
--   				- activate / deactivate increasing motor rpm for each action (only Admin can do this in multiplayer)
--   				- this can be set as global or for current vehicle type (it will be saved in config file for all savegames)
--   				- show / hide rpm HUD
--   				- show / hide help in F1 menu
--   				- set custom size of rpm HUD
-- 				- rpm increase to these values:
--   				- movingTools - minMotorRpm + 15%
--   				- tipping - minMotorRpm + 40%
--   				- folding / unfolding - minMotorRpm + 30%
--   				- lifting / rotating implement - minMotorRpm + 30%
--   				- folding / unfolding pipe - minMotorRpm + 10%
-- 				- rpm color switches to red when rpm reach 80% of maxMotorRpm
-- 				- all controls can be set in your Inputbindings
-- @author:    	kenny456 (kenny456@seznam.cz)
-- @history:	v1.0.0.0 - 2020-11-11 - first release
--
WorkRPM = {}
WorkRPM.confDir = getUserProfileAppPath().. "modsSettings/WorkRPM/"
WorkRPM.modDirectory = g_currentModDirectory

WorkRPM.booleans = { true, false }
WorkRPM.hudSizes = { 0.8, 1, 1.2, 1.4, 0 }
WorkRPM.hudSizesCustom = { 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1, 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 1.9, 2, 2.1, 2.2, 2.3, 2.4, 2.5, 2.6, 2.7, 2.8, 2.9, 3 }

WorkRPM.booleansTexts = { 'On', 'Off' }
WorkRPM.hudSizesTexts = { g_i18n:getText('WORKRPM_HUD_SMALL'), g_i18n:getText('WORKRPM_HUD_NORMAL'), g_i18n:getText('WORKRPM_HUD_LARGE'), g_i18n:getText('WORKRPM_HUD_LARGER'), g_i18n:getText('WORKRPM_HUD_CUSTOM') }
WorkRPM.hudSizesCustomTexts = { '0.1', '0.2', '0.3', '0.4', '0.5', '0.6', '0.7', '0.8', '0.9', '1', '1.1', '1.2', '1.3', '1.4', '1.5', '1.6', '1.7', '1.8', '1.9', '2', '2.1', '2.2', '2.3', '2.4', '2.5', '2.6', '2.7', '2.8', '2.9', '3' }



WorkRPM.allowTipping = true
WorkRPM.allowFolding = true
WorkRPM.allowRotating = true
WorkRPM.allowLifting = true
WorkRPM.allowMovingTools = true
WorkRPM.allowPipeFolding = true
WorkRPM.showHelp = true
WorkRPM.showHud = true
WorkRPM.hudSize = 1
WorkRPM.hudSizeCustom = 1

WorkRPM.maxSize = 1.4
WorkRPM.minSize = 0.8
WorkRPM.shiftX = 0.022
WorkRPM.shiftY = 0.85


WorkRPM.allowTippingIndex = 1
WorkRPM.allowFoldingIndex = 1
WorkRPM.allowRotatingIndex = 1
WorkRPM.allowLiftingIndex = 1
WorkRPM.allowMovingToolsIndex = 1
WorkRPM.allowPipeFoldingIndex = 1
WorkRPM.showHelpIndex = 1
WorkRPM.showHudIndex = 1
WorkRPM.hudSizeIndex = 2
WorkRPM.hudSizeCustomIndex = 10

function getIndexFromElement(table, element)
	for index, el in pairs(table) do
		if type(element) == 'number' then
			if string.format("%.2f", element) == string.format("%.2f", el) then
				return index
			end
		else
			if element == el then
				return index
			end
		end
	end
	return false
end
function WorkRPM:settingsFromGui(vehicle, tippingState, foldingState, rotatingState, liftingState, movingToolsState, pipeState, 
											currentConfigState, tippingStateCurrent, foldingStateCurrent, rotatingStateCurrent, liftingStateCurrent, movingToolsStateCurrent, pipeStateCurrent,
											showhelpState, showHudState, hudSizeState, hudSizeCustomState)
	local spec = vehicle.spec_workRPM
	if WorkRPM.showHelpIndex ~= showhelpState or WorkRPM.showHudIndex ~= showHudState or WorkRPM.hudSizeIndex ~= hudSizeState or WorkRPM.hudSizeCustomIndex ~= hudSizeCustomState then
		WorkRPM.showHelp = WorkRPM.booleans[showhelpState]
		WorkRPM.showHud = WorkRPM.booleans[showHudState]
		WorkRPM.hudSize = WorkRPM.hudSizes[hudSizeState]
		WorkRPM.hudSizeCustom = WorkRPM.hudSizesCustom[hudSizeCustomState]
		WorkRPM.showHelpIndex = showhelpState
		WorkRPM.showHudIndex = showHudState
		WorkRPM.hudSizeIndex = hudSizeState
		WorkRPM.hudSizeCustomIndex = hudSizeCustomState
		vehicle:saveToXmlWorkRPM(self)
		vehicle:updateRpmHud()
	end
	if WorkRPM.allowTippingIndex ~= tippingState or WorkRPM.allowFoldingIndex ~= foldingState or WorkRPM.allowRotatingIndex ~= rotatingState or WorkRPM.allowLiftingIndex ~= liftingState
	or WorkRPM.allowMovingToolsIndex ~= movingToolsState or WorkRPM.allowPipeFoldingIndex ~= pipeState then
		vehicle:wrpmSetGlobalConfig(WorkRPM.booleans[tippingState], WorkRPM.booleans[foldingState], WorkRPM.booleans[rotatingState], WorkRPM.booleans[liftingState], WorkRPM.booleans[movingToolsState], WorkRPM.booleans[pipeState], false)
	end
	
	if spec.currentConfigIndex ~= currentConfigState or spec.allowTippingIndex ~= tippingStateCurrent or spec.allowFoldingIndex ~= foldingStateCurrent or spec.allowRotatingIndex ~= rotatingStateCurrent or spec.allowLiftingIndex ~= liftingStateCurrent
	or spec.allowMovingToolsIndex ~= movingToolsStateCurrent or spec.allowPipeFoldingIndex ~= pipeStateCurrent then
		vehicle:wrpmSetVehicleConfig(WorkRPM.booleans[currentConfigState], WorkRPM.booleans[tippingStateCurrent], WorkRPM.booleans[foldingStateCurrent], WorkRPM.booleans[rotatingStateCurrent], WorkRPM.booleans[liftingStateCurrent], WorkRPM.booleans[movingToolsStateCurrent],
		WorkRPM.booleans[pipeStateCurrent], false)
	end
end
function WorkRPM:settingsResetGui()
	if g_currentMission.isMasterUser then
		WorkRPM.gui.wrpmSettingGui:setTipping(1)
		WorkRPM.gui.wrpmSettingGui:setFolding(1)
		WorkRPM.gui.wrpmSettingGui:setRotating(1)
		WorkRPM.gui.wrpmSettingGui:setLifting(1)
		WorkRPM.gui.wrpmSettingGui:setMovingTools(1)
		WorkRPM.gui.wrpmSettingGui:setPipe(1)
		
		WorkRPM.gui.wrpmSettingGui:setCurrentConfig(2)
		WorkRPM.gui.wrpmSettingGui:setTippingCurrent(1)
		WorkRPM.gui.wrpmSettingGui:setFoldingCurrent(1)
		WorkRPM.gui.wrpmSettingGui:setRotatingCurrent(1)
		WorkRPM.gui.wrpmSettingGui:setLiftingCurrent(1)
		WorkRPM.gui.wrpmSettingGui:setMovingToolsCurrent(1)
		WorkRPM.gui.wrpmSettingGui:setPipeCurrent(1)
		
		WorkRPM.gui.wrpmSettingGui.tippingCurrent:setDisabled(true)
		WorkRPM.gui.wrpmSettingGui.foldingCurrent:setDisabled(true)
		WorkRPM.gui.wrpmSettingGui.rotatingCurrent:setDisabled(true)
		WorkRPM.gui.wrpmSettingGui.liftingCurrent:setDisabled(true)
		WorkRPM.gui.wrpmSettingGui.movingToolsCurrent:setDisabled(true)
		WorkRPM.gui.wrpmSettingGui.pipeCurrent:setDisabled(true)
	end
	WorkRPM.gui.wrpmSettingGui:setShowhelp(1)
	WorkRPM.gui.wrpmSettingGui:setShowHud(1)
	WorkRPM.gui.wrpmSettingGui:setHudSize(2)
	WorkRPM.gui.wrpmSettingGui:setHudSizeCustom(10)
	WorkRPM.gui.wrpmSettingGui.hudSizeCustom:setDisabled(true)
end
function WorkRPM.prerequisitesPresent(specializations)
	return true
end
function WorkRPM.registerOverwrittenFunctions(vehicleType)
end
function WorkRPM.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "saveToXmlWorkRPM", 			WorkRPM.saveToXmlWorkRPM)
	SpecializationUtil.registerFunction(vehicleType, "loadFromXmlWorkRPM", 			WorkRPM.loadFromXmlWorkRPM)
	SpecializationUtil.registerFunction(vehicleType, "wrpmGetFullName", 			WorkRPM.wrpmGetFullName)
	SpecializationUtil.registerFunction(vehicleType, "wrpmSetGlobalConfig", 			WorkRPM.wrpmSetGlobalConfig)
	SpecializationUtil.registerFunction(vehicleType, "wrpmSetVehicleConfig", 			WorkRPM.wrpmSetVehicleConfig)
	SpecializationUtil.registerFunction(vehicleType, "updateRpmHud", 			WorkRPM.updateRpmHud)
end
function WorkRPM:onRegisterActionEvents(isSelected, isOnActiveVehicle)
	local spec = self.spec_workRPM
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	if spec.event_IDs == nil then
		spec.event_IDs = {}
	else
		self:clearActionEventsTable(spec.event_IDs)
	end
	if self.getIsControlled ~= nil and self:getIsControlled() then
		self:loadFromXmlWorkRPM(self:wrpmGetFullName())
		if g_dedicatedServerInfo ~= nil then
			return
		end
		self:updateRpmHud()
		local actions = { InputAction.WORKRPM_OPEN_MENU, InputAction.WORKRPM_MOVE_HUD_TRIGGER, InputAction.WORKRPM_MOVE_HUD_LEFT, InputAction.WORKRPM_MOVE_HUD_RIGHT, InputAction.WORKRPM_MOVE_HUD_UP, InputAction.WORKRPM_MOVE_HUD_DOWN,
							InputAction.WORKRPM_TOGGLE_SIZE }
		for _,actionName in pairs(actions) do
			local always = (actionName == InputAction.WORKRPM_MOVE_HUD_TRIGGER or actionName == InputAction.WORKRPM_MOVE_HUD_LEFT or actionName == InputAction.WORKRPM_MOVE_HUD_RIGHT or actionName == InputAction.WORKRPM_MOVE_HUD_UP or actionName == InputAction.WORKRPM_MOVE_HUD_DOWN) and true or false
			local _, eventID = g_inputBinding:registerActionEvent(actionName, self, WorkRPM.actionCallback, true, true, always, true)
			spec.event_IDs[actionName] = eventID
			if g_inputBinding ~= nil and g_inputBinding.events ~= nil and g_inputBinding.events[eventID] ~= nil then
				if actionName == 'something with lower priority' then
					g_inputBinding:setActionEventTextPriority(eventID, GS_PRIO_NORMAL)
				else
					g_inputBinding:setActionEventTextPriority(eventID, GS_PRIO_VERY_HIGH)
				end
				g_inputBinding:setActionEventTextVisibility(eventID, WorkRPM.showHelp)
			end
			local colliding = false
			_, colliding, _ = g_inputBinding:checkEventCollision(actionName)
			if colliding then
				if g_inputBinding.nameActions[actionName].bindings[1] ~= nil then
					if g_inputBinding.nameActions[actionName].bindings[1].inputString ~= nil then
						print(string.format('Warning: WorkRPM got a colliding input action: %s', actionName)..' ('..g_inputBinding.nameActions[actionName].bindings[1].inputString..'). You can remap it in controls settings')
					end
				else
					print(string.format('Warning: WorkRPM got a colliding input action: %s', actionName))
				end
			end
		end
	end
end
function WorkRPM.registerEventListeners(vehicleType)
	for _,n in pairs( { "onLoad", "onPostLoad", "saveToXMLFile", "onUpdate", "onUpdateTick", "onDraw", "onReadStream", "onWriteStream", "onRegisterActionEvents", "wrpmSetGlobalConfig", "wrpmSetVehicleConfig" } ) do
		SpecializationUtil.registerEventListener(vehicleType, n, WorkRPM)
	end
end
function WorkRPM:onLoad(savegame)
	self.getPtoRpm = Utils.overwrittenFunction(self.getPtoRpm, WorkRPM.getPtoRpm)
	--self.getDoConsumePtoPower = Utils.overwrittenFunction(self.getDoConsumePtoPower, WorkRPM.getDoConsumePtoPower)
	if self.isClient then
		if WorkRPM.gui == nil then
			g_gui:loadProfiles(WorkRPM.modDirectory .. "gui/guiProfiles.xml")
			WorkRPM.gui = {}
			WorkRPM.gui["wrpmSettingGui"] = workRPMGui:new()
			g_gui:loadGui(WorkRPM.modDirectory .. "gui/workRPMGui.xml", "workRPMGui", WorkRPM.gui.wrpmSettingGui);
		end
	end
	
	self.spec_workRPM = {}
	local spec = self.spec_workRPM
	
	spec.modInitialized = true
	spec.modAllowed = true
	
	local xmlFile = self.xmlFile
	spec.rpm = 0
	spec.releaseRpmTimer = 0
	spec.maxTime = 50
	spec.minRpm = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.motorized.motorConfigurations.motorConfiguration.motor#minRpm"), 850)
	spec.maxRpm = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.motorized.motorConfigurations.motorConfiguration.motor#maxRpm"), 2200)
	
	if hasXMLProperty(xmlFile, 'vehicle.WorkRPM') then
		spec.rpmMovingTools = Utils.getNoNil(getXMLInt(xmlFile, "vehicle.WorkRPM#rpmMovingTools"), spec.minRpm * 1.15) / 4
		spec.rpmTipping = Utils.getNoNil(getXMLInt(xmlFile, "vehicle.WorkRPM#rpmTipping"), spec.minRpm * 1.4) / 4
		spec.rpmFolding = Utils.getNoNil(getXMLInt(xmlFile, "vehicle.WorkRPM#rpmFolding"), spec.minRpm * 1.3) / 4
		spec.rpmPipeFolding = Utils.getNoNil(getXMLInt(xmlFile, "vehicle.WorkRPM#rpmPipeFolding"), spec.minRpm * 1.1) / 4
		spec.rpmLifting = Utils.getNoNil(getXMLInt(xmlFile, "vehicle.WorkRPM#rpmLifting"), spec.minRpm * 1.3) / 4
		spec.rpmRotating = Utils.getNoNil(getXMLInt(xmlFile, "vehicle.WorkRPM#rpmRotating"), spec.minRpm * 1.3) / 4
	else
		spec.rpmMovingTools = (spec.minRpm * 1.15) / 4
		spec.rpmTipping = (spec.minRpm * 1.4) / 4	
		spec.rpmFolding = (spec.minRpm * 1.3) / 4
		spec.rpmPipeFolding = (spec.minRpm * 1.1) / 4
		spec.rpmLifting = (spec.minRpm * 1.3) / 4
		spec.rpmRotating = (spec.minRpm * 1.3) / 4
	end
	
	spec.overlay = createImageOverlay(WorkRPM.modDirectory .. "overlay_bg.dds")
    setOverlayColor(spec.overlay, 0, 0, 0, 0.75)
	
    spec.neededRpmTemp = spec.minRpm / 4
	spec.neededRpmBack = 0
	spec.inputLowerImplement = false
	spec.isFoldable = false
	spec.foldableAnim = nil
	spec.isLifting = false
	spec.liftTimer = 0
	spec.lowerInput = false
	spec.lowerInputTimer = 0
	spec.lowerInputReleased = true
	spec.timerHud = 0
	
	spec.currentConfig = false
	spec.allowFolding = true
	spec.allowTipping = true
	spec.allowRotating = true
	spec.allowLifting = true
	spec.allowMovingTools = true
	spec.allowPipeFolding = true
	
	spec.currentConfigIndex = 2
	spec.allowTippingIndex = 1
	spec.allowFoldingIndex = 1
	spec.allowRotatingIndex = 1
	spec.allowLiftingIndex = 1
	spec.allowMovingToolsIndex = 1
	spec.allowPipeFoldingIndex = 1
	
	spec.ui = g_gameSettings.uiScale
	
	if self.spec_pipe ~= nil and SpecializationUtil.hasSpecialization(Pipe, self.specializations)then
		spec.isOverloadVehicle = true
	end
	
	if SpecializationUtil.hasSpecialization(Foldable, self.specializations) then
		local anim = getXMLString(self.xmlFile, "vehicle.foldable.foldingParts.foldingPart#animationName")
		if anim ~= nil then
			spec.isFoldable = true
			spec.foldableAnim = anim
		end
	end
	local configFile = WorkRPM.confDir .. "WorkRPMConfig.xml"
	if fileExists(configFile) then
		self:loadFromXmlWorkRPM(self:wrpmGetFullName())
	else
		createFolder(getUserProfileAppPath().. "modsSettings/")
		createFolder(WorkRPM.confDir)
		WorkRPM.configXml = createXMLFile("WorkRPM_XML", configFile, "WorkRPMConfig")
		self:saveToXmlWorkRPM(self)
	end
end
function WorkRPM:onPostLoad(savegame)
	local spec = self.spec_workRPM
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	self:updateRpmHud()
end
function WorkRPM:updateRpmHud()
	local spec = self.spec_workRPM
	
	spec.sizeHud = (WorkRPM.hudSizeIndex == 5 and WorkRPM.hudSizeCustom or WorkRPM.hudSize) * spec.ui
	spec.hudPositionX = g_currentMission.inGameMenu.hud.speedMeter.gaugeCenterX - WorkRPM.shiftX
	spec.hudPositionY = g_currentMission.inGameMenu.hud.speedMeter.gaugeCenterX - WorkRPM.shiftY
	spec.hudWidth = 0.045 * spec.sizeHud
	spec.hudHeight = 0.017 * spec.sizeHud
	spec.hudTextX = spec.hudPositionX + (0.023 * spec.sizeHud)
	spec.hudTextY = spec.hudPositionY + (0.018 * spec.sizeHud)
end
function WorkRPM:loadFromXmlWorkRPM(vehicle)
	local spec = self.spec_workRPM
	
	local configFile = WorkRPM.confDir .. "WorkRPMConfig.xml"
	if fileExists(configFile) then
		WorkRPM.configXml = loadXMLFile("WorkRPM_XML", configFile)
	end
	if fileExists(configFile) then
		if getXMLBool(WorkRPM.configXml, "WorkRPMConfig.showHelp") ~= nil then
			WorkRPM.showHelp = getXMLBool(WorkRPM.configXml, "WorkRPMConfig.showHelp")
		end
		if getXMLBool(WorkRPM.configXml, "WorkRPMConfig.showHud") ~= nil then
			WorkRPM.showHud = getXMLBool(WorkRPM.configXml, "WorkRPMConfig.showHud")
		end
		if getXMLFloat(WorkRPM.configXml, "WorkRPMConfig.hudSize") ~= nil then
			WorkRPM.hudSize = getXMLFloat(WorkRPM.configXml, "WorkRPMConfig.hudSize")
		end
		if getXMLFloat(WorkRPM.configXml, "WorkRPMConfig.hudSizeCustom") ~= nil then
			WorkRPM.hudSizeCustom = getXMLFloat(WorkRPM.configXml, "WorkRPMConfig.hudSizeCustom")
		end
		if getXMLFloat(WorkRPM.configXml, "WorkRPMConfig.shiftX") ~= nil then
			WorkRPM.shiftX = getXMLFloat(WorkRPM.configXml, "WorkRPMConfig.shiftX")
		end
		if getXMLFloat(WorkRPM.configXml, "WorkRPMConfig.shiftY") ~= nil then
			WorkRPM.shiftY = getXMLFloat(WorkRPM.configXml, "WorkRPMConfig.shiftY")
		end
		if getXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowTipping") ~= nil then
			WorkRPM.allowTipping = getXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowTipping")
		end
		if getXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowFolding") ~= nil then
			WorkRPM.allowFolding = getXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowFolding")
		end
		if getXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowRotating") ~= nil then
			WorkRPM.allowRotating = getXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowRotating")
		end
		if getXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowLifting") ~= nil then
			WorkRPM.allowLifting = getXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowLifting")
		end
		if getXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowMovingTools") ~= nil then
			WorkRPM.allowMovingTools = getXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowMovingTools")
		end
		if getXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowPipeFolding") ~= nil then
			WorkRPM.allowPipeFolding = getXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowPipeFolding")
		end
		if vehicle ~= nil then
			local key = "WorkRPMConfig.wrpmActiveVehicles"
			local i = 0
			while true do
				local name = getXMLString(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#name")
				if name ~= nil then
					if name == vehicle then
						spec.currentConfig = getXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#currentConfig")
						spec.allowTipping = getXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowTipping")
						spec.allowFolding = getXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowFolding")
						spec.allowRotating = getXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowRotating")
						spec.allowLifting = getXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowLifting")
						spec.allowMovingTools = getXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowMovingTools")
						spec.allowPipeFolding = getXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowPipeFolding")
						
						spec.currentConfigIndex = getIndexFromElement(WorkRPM.booleans, spec.currentConfig)
						spec.allowTippingIndex = getIndexFromElement(WorkRPM.booleans, spec.allowTipping)
						spec.allowFoldingIndex = getIndexFromElement(WorkRPM.booleans, spec.allowFolding)
						spec.allowRotatingIndex = getIndexFromElement(WorkRPM.booleans, spec.allowRotating)
						spec.allowLiftingIndex = getIndexFromElement(WorkRPM.booleans, spec.allowLifting)
						spec.allowMovingToolsIndex = getIndexFromElement(WorkRPM.booleans, spec.allowMovingTools)
						spec.allowPipeFoldingIndex = getIndexFromElement(WorkRPM.booleans, spec.allowPipeFolding)
						break
					end
				else
					break
				end
				i = i + 1
			end
		end
		WorkRPM.allowTippingIndex = getIndexFromElement(WorkRPM.booleans, WorkRPM.allowTipping)
		WorkRPM.allowFoldingIndex = getIndexFromElement(WorkRPM.booleans, WorkRPM.allowFolding)
		WorkRPM.allowRotatingIndex = getIndexFromElement(WorkRPM.booleans, WorkRPM.allowRotating)
		WorkRPM.allowLiftingIndex = getIndexFromElement(WorkRPM.booleans, WorkRPM.allowLifting)
		WorkRPM.allowMovingToolsIndex = getIndexFromElement(WorkRPM.booleans, WorkRPM.allowMovingTools)
		WorkRPM.allowPipeFoldingIndex = getIndexFromElement(WorkRPM.booleans, WorkRPM.allowPipeFolding)
		
		WorkRPM.showHelpIndex = getIndexFromElement(WorkRPM.booleans, WorkRPM.showHelp)
		WorkRPM.showHudIndex = getIndexFromElement(WorkRPM.booleans, WorkRPM.showHud)
		WorkRPM.hudSizeIndex = getIndexFromElement(WorkRPM.hudSizes, WorkRPM.hudSize)
		WorkRPM.hudSizeCustomIndex = getIndexFromElement(WorkRPM.hudSizesCustom, WorkRPM.hudSizeCustom)
	end
end
function WorkRPM:saveToXmlWorkRPM(vehicleId, vehicle)
	local spec = vehicleId.spec_workRPM
	
	if WorkRPM.configXml ~= nil then
		setXMLBool(WorkRPM.configXml, "WorkRPMConfig.showHelp", WorkRPM.showHelp)
		setXMLBool(WorkRPM.configXml, "WorkRPMConfig.showHud", WorkRPM.showHud)
		setXMLFloat(WorkRPM.configXml, "WorkRPMConfig.hudSize", WorkRPM.hudSize)
		setXMLFloat(WorkRPM.configXml, "WorkRPMConfig.hudSizeCustom", WorkRPM.hudSizeCustom)
		setXMLFloat(WorkRPM.configXml, "WorkRPMConfig.shiftX", WorkRPM.shiftX)
		setXMLFloat(WorkRPM.configXml, "WorkRPMConfig.shiftY", WorkRPM.shiftY)
		setXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowTipping", WorkRPM.allowTipping)
		setXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowFolding", WorkRPM.allowFolding)
		setXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowRotating", WorkRPM.allowRotating)
		setXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowLifting", WorkRPM.allowLifting)
		setXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowMovingTools", WorkRPM.allowMovingTools)
		setXMLBool(WorkRPM.configXml, "WorkRPMConfig.allowPipeFolding", WorkRPM.allowPipeFolding)
		if vehicle ~= nil then
			local key = "WorkRPMConfig.wrpmActiveVehicles"
			local i = 0
			while true do
				local name = getXMLString(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#name")
				if name ~= nil then
					if name == vehicle then
						setXMLString(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#name", vehicle)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#currentConfig", spec.currentConfig)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowTipping", spec.allowTipping)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowFolding", spec.allowFolding)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowRotating", spec.allowRotating)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowLifting", spec.allowLifting)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowMovingTools", spec.allowMovingTools)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowPipeFolding", spec.allowPipeFolding)
						break
					end
				else
					setXMLString(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#name", vehicle)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#currentConfig", spec.currentConfig)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowTipping", spec.allowTipping)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowFolding", spec.allowFolding)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowRotating", spec.allowRotating)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowLifting", spec.allowLifting)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowMovingTools", spec.allowMovingTools)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowPipeFolding", spec.allowPipeFolding)
					break
				end
				i = i + 1
				local name = key .. string.format(".vehicle(%d)", i)
				if not hasXMLProperty(WorkRPM.configXml, name) then
					setXMLString(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#name", vehicle)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#currentConfig", spec.currentConfig)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowTipping", spec.allowTipping)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowFolding", spec.allowFolding)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowRotating", spec.allowRotating)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowLifting", spec.allowLifting)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowMovingTools", spec.allowMovingTools)
						setXMLBool(WorkRPM.configXml, "WorkRPMConfig.wrpmActiveVehicles"..string.format(".vehicle(%d)", i).."#allowPipeFolding", spec.allowPipeFolding)
					break
				end
			end
		end
		saveXMLFile(WorkRPM.configXml)
	end
end
function WorkRPM:saveToXMLFile(xmlFile, key)
	local spec = self.spec_workRPM
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
end
function WorkRPM:onUpdate(dt)
	local spec = self.spec_workRPM
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	local debug = false
	if g_inputBinding.nameActions['LOWER_IMPLEMENT'] ~= nil and g_inputBinding.nameActions['LOWER_IMPLEMENT'].activeBindings ~= nil then
		for _,activeBinding in pairs(g_inputBinding.nameActions['LOWER_IMPLEMENT'].activeBindings) do
			if activeBinding.isDown then
				spec.inputLowerImplement = true
				WorkRPMLowerImplementEvent.sendEvent(self, false)
			end
		end
	end
	if self:getIsActive() then
		local allowTipping, allowFolding, allowRotating, allowLifting, allowMovingTools, allowPipeFolding
		if spec.currentConfig then
			allowTipping = spec.allowTipping
			allowFolding = spec.allowFolding
			allowRotating = spec.allowRotating
			allowLifting = spec.allowLifting
			allowMovingTools = spec.allowMovingTools
			allowPipeFolding = spec.allowPipeFolding
		else
			allowTipping = WorkRPM.allowTipping
			allowFolding = WorkRPM.allowFolding
			allowRotating = WorkRPM.allowRotating
			allowLifting = WorkRPM.allowLifting
			allowMovingTools = WorkRPM.allowMovingTools
			allowPipeFolding = WorkRPM.allowPipeFolding
		end
		local neededRpm = 0
		--		Tipping		--
		if neededRpm == 0 and allowTipping then
			if spec.isOverloadVehicle then
				if self.spec_dischargeable ~= nil then
					if self.spec_dischargeable.currentDischargeState == 1 or self.spec_dischargeable.currentDischargeState == 2 then
						neededRpm = spec.rpmTipping
						if debug then print('tipping1') end
					end
				end
			else
				if self.spec_trailer ~= nil then
					if self.spec_trailer.tipState == Trailer.TIPSTATE_OPENING then
						neededRpm = spec.rpmTipping
						if debug then print('tipping2') end
					end
				end
			end
			if self.spec_attacherJoints ~= nil then
				local trailer
				for _,implement in pairs(self.spec_attacherJoints.attachedImplements) do
					if implement.object ~= nil then
						trailer = implement.object
						local i = 1
						while true do
							if trailer.spec_trailer ~= nil then
								if trailer.spec_pipe ~= nil and SpecializationUtil.hasSpecialization(Pipe, trailer.specializations)then
									if trailer.spec_dischargeable.currentDischargeState == 1 or trailer.spec_dischargeable.currentDischargeState == 2 then
										neededRpm = spec.rpmTipping
										if debug then print('tipping trailer1') end
										break
									end
								else
									if trailer.spec_trailer.tipState == Trailer.TIPSTATE_OPENING then
										neededRpm = spec.rpmTipping
										if debug then print('tipping trailer2') end
										break
									end
								end
							end
							if trailer.spec_attacherJoints ~= nil then
								if trailer.spec_attacherJoints.attachedImplements ~= nil then
									if trailer.spec_attacherJoints.attachedImplements[1] ~= nil and  trailer.spec_attacherJoints.attachedImplements[1].object ~= nil then
										trailer = trailer.spec_attacherJoints.attachedImplements[1].object
									else
										break
									end
								else
									break
								end
							else
								break
							end
							i = i + 1
							if i > 10 then break end
						end
					end
				end
			end
		end
		--		Folding		--
		if neededRpm == 0 and allowFolding then
			if spec.isFoldable == true and spec.foldableAnim ~= nil then
				if self:getIsAnimationPlaying(spec.foldableAnim) then
					neededRpm = spec.rpmFolding
					if debug then print('folding') end
				end
			end
			if self.spec_attacherJoints ~= nil then
				local trailer
				for _,implement in pairs(self.spec_attacherJoints.attachedImplements) do
					if implement.object ~= nil then
						trailer = implement.object
						local i = 1
						while true do
							if SpecializationUtil.hasSpecialization(Foldable, trailer.specializations) then
								local anim = getXMLString(trailer.xmlFile, "vehicle.foldable.foldingParts.foldingPart#animationName");
								if anim ~= nil then
									if trailer:getIsAnimationPlaying(anim) then
										trailer:getAnimationTime(ianim)
										neededRpm = spec.rpmFolding
										if debug then print(i..' folding') end
										break
									end
								end
							end
							if trailer.spec_attacherJoints ~= nil then
								if trailer.spec_attacherJoints.attachedImplements ~= nil then
									if trailer.spec_attacherJoints.attachedImplements[1] ~= nil and  trailer.spec_attacherJoints.attachedImplements[1].object ~= nil then
										trailer = trailer.spec_attacherJoints.attachedImplements[1].object
									else
										break
									end
								else
									break
								end
							else
								break
							end
							i = i + 1
							if i > 10 then break end
						end
					end
				end
			end
		end
		--		implements rotation		--
		if neededRpm == 0 and allowRotating then
			if self.spec_attacherJoints ~= nil then
				local trailer
				for _,implement in pairs(self.spec_attacherJoints.attachedImplements) do
					if implement.object ~= nil then
						trailer = implement.object
						local i = 1
						while true do
							if trailer.spec_plow ~= nil then
								if trailer.spec_plow.rotationPart ~= nil then
									if trailer.spec_plow.rotationPart.turnAnimation ~= nil and trailer.getAnimationTime ~= nil then
										local rotationTime = trailer:getAnimationTime(trailer.spec_plow.rotationPart.turnAnimation);
										if rotationTime > 0 and rotationTime < 1 then
											neededRpm = spec.rpmRotating
											if debug then print(i..' rotating') end
											break
										end
									end
								end
							end
							if trailer.spec_attacherJoints ~= nil then
								if trailer.spec_attacherJoints.attachedImplements ~= nil then
									if trailer.spec_attacherJoints.attachedImplements[1] ~= nil and  trailer.spec_attacherJoints.attachedImplements[1].object ~= nil then
										trailer = trailer.spec_attacherJoints.attachedImplements[1].object
									else
										break
									end
								else
									break
								end
							else
								break
							end
							i = i + 1
							if i > 10 then break end
						end
					end
				end
			end
		end
		--		implements lowering / lifting		--
		if neededRpm == 0 and allowLifting then
			if self.spec_attacherJoints ~= nil then
				local trailer
				for _,implement in pairs(self.spec_attacherJoints.attachedImplements) do
					if implement.object ~= nil then
						trailer = implement.object
						local i = 1
						while true do
							if trailer.spec_attachable.attacherVehicle ~= nil and trailer.spec_attachable.attacherVehicle.spec_attacherJoints.selectedImplement ~= nil and trailer.spec_attachable.attacherVehicle.spec_attacherJoints.attacherJoints ~= nil then
								local jointDesc = trailer.spec_attachable.attacherVehicle.spec_attacherJoints.attacherJoints[trailer.spec_attachable.attacherVehicle.spec_attacherJoints.selectedImplement.jointDescIndex];
								if self:getIsActive() then
									if spec.inputLowerImplement then
										spec.inputLowerImplement = false
										if spec.lowerInputReleased == true then
											spec.lowerInput = true
											spec.lowerInputReleased = false
										end
									else
										spec.lowerInputReleased = true
									end
									if spec.lowerInput == true then
										spec.lowerInputTimer = spec.lowerInputTimer + dt
										if spec.lowerInputTimer > 200 then
											spec.lowerInputTimer = 0
											spec.lowerInput = false
											if spec.jointDown == true then
												spec.isLifting = true
												spec.jointDown = false
											else
												spec.jointDown = jointDesc.moveDown
											end
										end
									else
										spec.jointDown = jointDesc.moveDown
									end
									if spec.isLifting == true then
										spec.liftTimer = spec.liftTimer + dt
										if spec.liftTimer < 1400 then
											neededRpm = spec.rpmLifting
											if debug then print(i..' lowering') end
										else
											spec.liftTimer = 0
											spec.isLifting = false
										end
									end
								end
							end
							if trailer.spec_attacherJoints ~= nil then
								if trailer.spec_attacherJoints.attachedImplements ~= nil then
									if trailer.spec_attacherJoints.attachedImplements[1] ~= nil and  trailer.spec_attacherJoints.attachedImplements[1].object ~= nil then
										trailer = trailer.spec_attacherJoints.attachedImplements[1].object
									else
										break
									end
								else
									break
								end
							else
								break
							end
							i = i + 1
							if i > 10 then break end
						end
					end
				end
			end
		end
		spec.inputLowerImplement = false
		--		movingTools		--
		if neededRpm == 0 and allowMovingTools then
			local trailer = {}
			local vehicle = {}
			trailer.tipState = 0
			if self.spec_attacherJoints ~= nil then
				for _, attachedTrailer in pairs(self.spec_attacherJoints.attachedImplements) do
					if attachedTrailer.object ~= nil then
						if attachedTrailer.object.spec_trailer ~= nil then
							trailer = attachedTrailer.object.spec_trailer
						end
					end
				end
			end
			if self.spec_trailer ~= nil then
				vehicle = self.spec_trailer
			end
			local justFolding = false
			if spec.isFoldable == true and spec.foldableAnim ~= nil then
				if self:getIsAnimationPlaying(spec.foldableAnim) then--------------------------------try if anim isMovingTool
					justFolding = true
				end
			end
			if self.spec_cylindered ~= nil then
				if self.spec_cylindered.movingTools ~= 0 and self.spec_cylindered.movingTools ~= nil then
					for _, tool in pairs(self.spec_cylindered.movingTools) do
						if tool.rotSpeed ~= nil then
							if tool.lastRotHistoryRPM == nil then
								tool.lastRotHistoryRPM = tool.curRot[tool.rotationAxis]
							end
							if 0.00001 < math.abs(tool.lastRotHistoryRPM - tool.curRot[tool.rotationAxis]) and vehicle.tipState ~= 3 and not justFolding then
								tool.lastRotHistoryRPM = tool.curRot[tool.rotationAxis]
								neededRpm = spec.rpmMovingTools
								if debug then print('movingTools') end
								break
							end
						end;
						if tool.transSpeed ~= nil then
							if tool.lastTransHistoryRPM == nil then
								tool.lastTransHistoryRPM = tool.curTrans[tool.translationAxis]
							end;
							if 0.00001 < math.abs(tool.lastTransHistoryRPM - tool.curTrans[tool.translationAxis]) and vehicle.tipState ~= 3 and not justFolding then
								tool.lastTransHistoryRPM = tool.curTrans[tool.translationAxis]
								neededRpm = spec.rpmMovingTools
								if debug then print('movingTools') end
								break
							end
						end
					end
				end
			end
		end
		--		implements movingTools		--
		if neededRpm == 0 and allowMovingTools then
			if self.spec_attacherJoints ~= nil then
				local trailer
				for _,implement in pairs(self.spec_attacherJoints.attachedImplements) do
					if implement.object ~= nil then
						trailer = implement.object
						local i = 1
						while true do
							local trailer2 = {}
							if trailer.spec_trailer ~= nil then
								trailer2 = trailer.spec_trailer
							end
							if trailer.spec_cylindered ~= nil then
								if trailer.spec_cylindered.movingTools ~= 0 and trailer.spec_cylindered.movingTools ~= nil then
									local justFolding = false
									if SpecializationUtil.hasSpecialization(Foldable, trailer.specializations) then
										local anim = getXMLString(trailer.xmlFile, "vehicle.foldable.foldingParts.foldingPart#animationName")
										if anim ~= nil then
											if trailer:getIsAnimationPlaying(anim) then
												justFolding = true
											end
										end
									end
									for _, tool in pairs(trailer.spec_cylindered.movingTools) do
										if tool.rotSpeed ~= nil then
											if tool.lastRotHistoryRPM == nil then
												tool.lastRotHistoryRPM = tool.curRot[tool.rotationAxis]
											end
											if 0.00001 < math.abs(tool.lastRotHistoryRPM - tool.curRot[tool.rotationAxis]) and trailer2.tipState ~= 3 and not justFolding then
												tool.lastRotHistoryRPM = tool.curRot[tool.rotationAxis]
												neededRpm = spec.rpmMovingTools
												if debug then print(i..' implements movingTools') end
												break
											end
										end
										if tool.transSpeed ~= nil then
											if tool.lastTransHistoryRPM == nil then
												tool.lastTransHistoryRPM = tool.curTrans[tool.translationAxis]
											end
											
											if 0.00001 < math.abs(tool.lastTransHistoryRPM - tool.curTrans[tool.translationAxis]) and trailer2.tipState ~= 3 and not justFolding then
												tool.lastTransHistoryRPM = tool.curTrans[tool.translationAxis]
												neededRpm = spec.rpmMovingTools
												if debug then print(i..' implements movingTools') end
												break
											end
										end
									end
								end
							end
							if trailer.spec_attacherJoints ~= nil then
								if trailer.spec_attacherJoints.attachedImplements ~= nil then
									if trailer.spec_attacherJoints.attachedImplements[1] ~= nil and  trailer.spec_attacherJoints.attachedImplements[1].object ~= nil then
										trailer = trailer.spec_attacherJoints.attachedImplements[1].object
									else
										break
									end
								else
									break
								end
							else
								break
							end
							i = i + 1
							if i > 10 then break end
						end
					end
				end
			end
		end
		--		Pipe folding / unfolding		--
		if neededRpm == 0 and allowPipeFolding then
			if spec.isOverloadVehicle then
				if self.spec_pipe.currentState == 0 then
					neededRpm = spec.rpmPipeFolding
					if debug then print('pipe self') end
				end
			end
			if self.spec_attacherJoints ~= nil then
				local trailer
				for _,implement in pairs(self.spec_attacherJoints.attachedImplements) do
					if implement.object ~= nil then
						trailer = implement.object
						local i = 1
						while true do
							if trailer.spec_pipe ~= nil and SpecializationUtil.hasSpecialization(Pipe, trailer.specializations) then
								if trailer.spec_pipe.currentState == 0 then
									neededRpm = spec.rpmPipeFolding
									if debug then print(i..' pipe implement') end
								end
							end
							if trailer.spec_attacherJoints ~= nil then
								if trailer.spec_attacherJoints.attachedImplements ~= nil then
									if trailer.spec_attacherJoints.attachedImplements[1] ~= nil and  trailer.spec_attacherJoints.attachedImplements[1].object ~= nil then
										trailer = trailer.spec_attacherJoints.attachedImplements[1].object
									else
										break
									end
								else
									break
								end
							else
								break
							end
							i = i + 1
							if i > 10 then break end
						end
					end
				end
			end
		end
		if neededRpm > 0 and self:getLastSpeed() < 3 then
			if neededRpm * 4 > self.spec_motorized.motor.lastMotorRpm then
				if spec.neededRpmTemp < neededRpm then
					spec.neededRpmBack = neededRpm
					spec.neededRpmTemp = spec.neededRpmTemp + (spec.neededRpmBack / 488.75)
				else
					spec.neededRpmTemp = neededRpm
				end
				spec.rpm = spec.neededRpmTemp
				spec.releaseRpmTimer = g_currentMission.time + spec.maxTime;
			end
		else
			if g_currentMission.time > spec.releaseRpmTimer then
				if spec.neededRpmTemp > spec.minRpm / 4 then
					spec.neededRpmTemp = spec.neededRpmTemp - ((spec.neededRpmBack / 488.75) * 3)
					spec.rpm = spec.neededRpmTemp;
				else
					spec.neededRpmTemp = spec.minRpm / 4
					spec.rpm = 0
				end
				spec.releaseRpmTimer = 0
			end
		end
	end
	if self.isClient then
		if self.getIsControlled ~= nil and self:getIsControlled() then
			if spec.event_IDs ~= nil and g_dedicatedServerInfo == nil then
				for actionName,eventID in pairs(spec.event_IDs) do
					if actionName == InputAction.WORKRPM_OPEN_MENU then
						g_inputBinding:setActionEventTextVisibility(eventID, WorkRPM.showHelp)
					elseif actionName == InputAction.WORKRPM_MOVE_HUD_TRIGGER then
						g_inputBinding:setActionEventTextVisibility(eventID, not spec.moveHudTrigger and WorkRPM.showHelp)
					elseif actionName == InputAction.WORKRPM_MOVE_HUD_LEFT or actionName == InputAction.WORKRPM_MOVE_HUD_RIGHT or actionName == InputAction.WORKRPM_MOVE_HUD_UP or actionName == InputAction.WORKRPM_MOVE_HUD_DOWN then
						g_inputBinding:setActionEventActive(eventID, spec.moveHudTrigger)
						g_inputBinding:setActionEventTextVisibility(eventID, WorkRPM.showHelp)
					elseif actionName == InputAction.WORKRPM_TOGGLE_SIZE then
						g_inputBinding:setActionEventText(eventID, g_i18n:getText('input_WORKRPM_TOGGLE_SIZE')..' : '..WorkRPM.hudSizesTexts[WorkRPM.hudSizeIndex])
						g_inputBinding:setActionEventTextVisibility(eventID, WorkRPM.showHelp)
					end
				end
			end
		end
		if spec.inputLeft or spec.inputRight or spec.inputUp or spec.inputDown then
			local x = WorkRPM.shiftX
			local y = WorkRPM.shiftY
			spec.timerHud = spec.timerHud + dt
			if spec.inputLeft then
				if spec.timerHud > 1500 then
					x = x + 0.001
				else
					x = x + 0.0001
				end
			end
			if spec.inputRight then
				if spec.timerHud > 1500 then
					x = x - 0.001
				else
					x = x - 0.0001
				end
			end
			if spec.inputUp then
				if spec.timerHud > 1500 then
					y = y - 0.001
				else
					y = y - 0.0001
				end
			end
			if spec.inputDown then
				if spec.timerHud > 1500 then
					y = y + 0.001
				else
					y = y + 0.0001
				end
			end
			WorkRPM.shiftX = x
			WorkRPM.shiftY = y
			self:updateRpmHud()
		else
			spec.timerHud = 0
		end
	end
end
function WorkRPM:wrpmGetFullName()
    local name = self:getName()
    local storeItem = g_storeManager:getItemByXMLFilename(self.configFileName)
    if storeItem ~= nil then
        local brand = g_brandManager:getBrandByIndex(storeItem.brandIndex)
        if brand ~= nil then
            name = brand.title .. " " .. name
        end
    end
    return name
end
function WorkRPM:getPtoRpm(oldFunc)
	local spec = self.spec_workRPM
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	return spec.rpm
end;
function WorkRPM:getDoConsumePtoPower(oldFunc)
	local spec = self.spec_workRPM
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	return spec.rpm > 0
end;
function WorkRPM:onUpdateTick(dt)
end
function WorkRPM:actionCallback(actionName, keyStatus, arg4, arg5, arg6)
	local spec = self.spec_workRPM
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	if self:getIsActive() then
		if keyStatus > 0 then
			if actionName == 'WORKRPM_OPEN_MENU' then
				if WorkRPM.gui.wrpmSettingGui.isOpen then
					WorkRPM.gui.wrpmSettingGui:onClickBack()
				elseif g_gui.currentGui == nil then
					WorkRPM.gui.wrpmSettingGui:setTarget(self)
											
					WorkRPM.gui.wrpmSettingGui:setTipping(WorkRPM.allowTippingIndex)
					WorkRPM.gui.wrpmSettingGui:setFolding(WorkRPM.allowFoldingIndex)
					WorkRPM.gui.wrpmSettingGui:setRotating(WorkRPM.allowRotatingIndex)
					WorkRPM.gui.wrpmSettingGui:setLifting(WorkRPM.allowLiftingIndex)
					WorkRPM.gui.wrpmSettingGui:setMovingTools(WorkRPM.allowMovingToolsIndex)
					WorkRPM.gui.wrpmSettingGui:setPipe(WorkRPM.allowPipeFoldingIndex)
					
					WorkRPM.gui.wrpmSettingGui:setCurrentConfig(spec.currentConfigIndex)
					
					WorkRPM.gui.wrpmSettingGui:setTippingCurrent(spec.allowTippingIndex)
					WorkRPM.gui.wrpmSettingGui:setFoldingCurrent(spec.allowFoldingIndex)
					WorkRPM.gui.wrpmSettingGui:setRotatingCurrent(spec.allowRotatingIndex)
					WorkRPM.gui.wrpmSettingGui:setLiftingCurrent(spec.allowLiftingIndex)
					WorkRPM.gui.wrpmSettingGui:setMovingToolsCurrent(spec.allowMovingToolsIndex)
					WorkRPM.gui.wrpmSettingGui:setPipeCurrent(spec.allowPipeFoldingIndex)
					
					WorkRPM.gui.wrpmSettingGui:setShowhelp(WorkRPM.showHelpIndex)
					WorkRPM.gui.wrpmSettingGui:setShowHud(WorkRPM.showHudIndex)
					WorkRPM.gui.wrpmSettingGui:setHudSize(WorkRPM.hudSizeIndex)
					WorkRPM.gui.wrpmSettingGui:setHudSizeCustom(WorkRPM.hudSizeCustomIndex)
					
					WorkRPM.gui.wrpmSettingGui.tipping:setDisabled(not g_currentMission.isMasterUser)
					WorkRPM.gui.wrpmSettingGui.folding:setDisabled(not g_currentMission.isMasterUser)
					WorkRPM.gui.wrpmSettingGui.rotating:setDisabled(not g_currentMission.isMasterUser)
					WorkRPM.gui.wrpmSettingGui.lifting:setDisabled(not g_currentMission.isMasterUser)
					WorkRPM.gui.wrpmSettingGui.movingTools:setDisabled(not g_currentMission.isMasterUser)
					WorkRPM.gui.wrpmSettingGui.pipe:setDisabled(not g_currentMission.isMasterUser)
					
					WorkRPM.gui.wrpmSettingGui.currentConfig:setDisabled(not g_currentMission.isMasterUser)
					WorkRPM.gui.wrpmSettingGui.tippingCurrent:setDisabled(not spec.currentConfig or not g_currentMission.isMasterUser)
					WorkRPM.gui.wrpmSettingGui.foldingCurrent:setDisabled(not spec.currentConfig or not g_currentMission.isMasterUser)
					WorkRPM.gui.wrpmSettingGui.rotatingCurrent:setDisabled(not spec.currentConfig or not g_currentMission.isMasterUser)
					WorkRPM.gui.wrpmSettingGui.liftingCurrent:setDisabled(not spec.currentConfig or not g_currentMission.isMasterUser)
					WorkRPM.gui.wrpmSettingGui.movingToolsCurrent:setDisabled(not spec.currentConfig or not g_currentMission.isMasterUser)
					WorkRPM.gui.wrpmSettingGui.pipeCurrent:setDisabled(not spec.currentConfig or not g_currentMission.isMasterUser)
					local dis = WorkRPM.gui.wrpmSettingGui.hudSize:getState() == 5
					WorkRPM.gui.wrpmSettingGui.hudSizeCustom:setDisabled(not dis)
					g_gui:showGui("workRPMGui")
				end
			elseif actionName == 'WORKRPM_TOGGLE_SIZE' then
				if WorkRPM.hudSizeIndex < 4 then
					WorkRPM.hudSizeIndex = WorkRPM.hudSizeIndex + 1
				else
					WorkRPM.hudSizeIndex = 1
				end
				WorkRPM.hudSize = WorkRPM.hudSizes[WorkRPM.hudSizeIndex]
				self:saveToXmlWorkRPM(self)
				self:updateRpmHud()
			end
		end
		if actionName == 'WORKRPM_MOVE_HUD_TRIGGER' then
			spec.moveHudTrigger = keyStatus > 0
		elseif actionName == 'WORKRPM_MOVE_HUD_LEFT' then
			spec.inputLeft = keyStatus > 0
			if keyStatus == 0 then self:saveToXmlWorkRPM(self) end
		elseif actionName == 'WORKRPM_MOVE_HUD_RIGHT' then
			spec.inputRight = keyStatus > 0
			if keyStatus == 0 then self:saveToXmlWorkRPM(self) end
		elseif actionName == 'WORKRPM_MOVE_HUD_UP' then
			spec.inputUp = keyStatus > 0
			if keyStatus == 0 then self:saveToXmlWorkRPM(self) end
		elseif actionName == 'WORKRPM_MOVE_HUD_DOWN' then
			spec.inputDown = keyStatus > 0
			if keyStatus == 0 then self:saveToXmlWorkRPM(self) end
		else
			spec.moveHudTrigger = false
		end
	end
end
function WorkRPM:wrpmSetGlobalConfig(allowTipping, allowFolding, allowRotating, allowLifting, allowMovingTools, allowPipeFolding, noEventSend)
	local spec = self.spec_workRPM
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	WorkRPM.allowTipping = allowTipping
	WorkRPM.allowFolding = allowFolding
	WorkRPM.allowRotating = allowRotating
	WorkRPM.allowLifting = allowLifting
	WorkRPM.allowMovingTools = allowMovingTools
	WorkRPM.allowPipeFolding = allowPipeFolding
	
	WorkRPM.allowTippingIndex = getIndexFromElement(WorkRPM.booleans, WorkRPM.allowTipping)
	WorkRPM.allowFoldingIndex = getIndexFromElement(WorkRPM.booleans, WorkRPM.allowFolding)
	WorkRPM.allowRotatingIndex = getIndexFromElement(WorkRPM.booleans, WorkRPM.allowRotating)
	WorkRPM.allowLiftingIndex = getIndexFromElement(WorkRPM.booleans, WorkRPM.allowLifting)
	WorkRPM.allowMovingToolsIndex = getIndexFromElement(WorkRPM.booleans, WorkRPM.allowMovingTools)
	WorkRPM.allowPipeFoldingIndex = getIndexFromElement(WorkRPM.booleans, WorkRPM.allowPipeFolding)
	
	self:saveToXmlWorkRPM(self)
	WorkRPMSetGlobalConfigEvent.sendEvent(self, allowTipping, allowFolding, allowRotating, allowLifting, allowMovingTools, allowPipeFolding, noEventSend)
end
function WorkRPM:wrpmSetVehicleConfig(currentConfig, allowTipping, allowFolding, allowRotating, allowLifting, allowMovingTools, allowPipeFolding, noEventSend)
	local spec = self.spec_workRPM
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	spec.currentConfig = currentConfig
	spec.allowTipping = allowTipping
	spec.allowFolding = allowFolding
	spec.allowRotating = allowRotating
	spec.allowLifting = allowLifting
	spec.allowMovingTools = allowMovingTools
	spec.allowPipeFolding = allowPipeFolding
	
	spec.currentConfigIndex = getIndexFromElement(WorkRPM.booleans, spec.currentConfig)
	spec.allowTippingIndex = getIndexFromElement(WorkRPM.booleans, spec.allowTipping)
	spec.allowFoldingIndex = getIndexFromElement(WorkRPM.booleans, spec.allowFolding)
	spec.allowRotatingIndex = getIndexFromElement(WorkRPM.booleans, spec.allowRotating)
	spec.allowLiftingIndex = getIndexFromElement(WorkRPM.booleans, spec.allowLifting)
	spec.allowMovingToolsIndex = getIndexFromElement(WorkRPM.booleans, spec.allowMovingTools)
	spec.allowPipeFoldingIndex = getIndexFromElement(WorkRPM.booleans, spec.allowPipeFolding)
	
	self:saveToXmlWorkRPM(self, self:wrpmGetFullName())
	WorkRPMSetVehicleConfigEvent.sendEvent(self, currentConfig, allowTipping, allowFolding, allowRotating, allowLifting, allowMovingTools, allowPipeFolding, noEventSend)
end
function WorkRPM:onDraw()
	local spec = self.spec_workRPM
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	if self.isClient and self.getIsControlled ~= nil and self:getIsControlled() and WorkRPM.showHud then
		if self.spec_motorized ~= nil then
			local rpm_txt = "-- rpm"
			if self.spec_motorized.isMotorStarted == true then
				rpm_txt = string.format("%i rpm", self.spec_motorized.motor.lastMotorRpm)
			end
			if self.spec_motorized.motor.lastMotorRpm > (spec.maxRpm / 10) * 8 then
				setTextColor(1, 0, 0, 1)
			else
				setTextColor(0, 1, 0, 1)
			end
			setTextAlignment(RenderText.ALIGN_CENTER)
			setTextVerticalAlignment(RenderText.VERTICAL_ALIGN_TOP)
			setTextBold(true)
			renderOverlay(spec.overlay, spec.hudPositionX, spec.hudPositionY, spec.hudWidth, spec.hudHeight)
			renderText(spec.hudTextX, spec.hudTextY, 0.015 * spec.sizeHud, rpm_txt)
			
			setTextColor(1,1,1,1)
			setTextAlignment(RenderText.ALIGN_LEFT)
			setTextVerticalAlignment(RenderText.VERTICAL_ALIGN_BASELINE)
			setTextBold(false)
		end
	end
end
function WorkRPM:onReadStream(streamId, connection)
	local spec = self.spec_workRPM
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	local allowTipping = streamReadBool(streamId)
	local allowFolding = streamReadBool(streamId)
	local allowRotating = streamReadBool(streamId)
	local allowLifting = streamReadBool(streamId)
	local allowMovingTools = streamReadBool(streamId)
	local allowPipeFolding = streamReadBool(streamId)
	if allowTipping ~= nil and allowFolding ~= nil and allowRotating ~= nil and allowLifting ~= nil and allowMovingTools ~= nil and allowPipeFolding ~= nil then
		self:wrpmSetGlobalConfig(allowTipping, allowFolding, allowRotating, allowLifting, allowMovingTools, allowPipeFolding, true)
	end
	
	local currentConfig = streamReadBool(streamId)
	local allowTipping = streamReadBool(streamId)
	local allowFolding = streamReadBool(streamId)
	local allowRotating = streamReadBool(streamId)
	local allowLifting = streamReadBool(streamId)
	local allowMovingTools = streamReadBool(streamId)
	local allowPipeFolding = streamReadBool(streamId)
	if currentConfig ~= nil and allowTipping ~= nil and allowFolding ~= nil and allowRotating ~= nil and allowLifting ~= nil and allowMovingTools ~= nil and allowPipeFolding ~= nil then
		self:wrpmSetVehicleConfig(currentConfig, allowTipping, allowFolding, allowRotating, allowLifting, allowMovingTools, allowPipeFolding, true)
	end
end
function WorkRPM:onWriteStream(streamId, connection)
	local spec = self.spec_workRPM
	if not spec.modInitialized or not spec.modAllowed then
		return
	end
	
	streamWriteBool(streamId, WorkRPM.allowTipping)
	streamWriteBool(streamId, WorkRPM.allowFolding)
	streamWriteBool(streamId, WorkRPM.allowRotating)
	streamWriteBool(streamId, WorkRPM.allowLifting)
	streamWriteBool(streamId, WorkRPM.allowMovingTools)
	streamWriteBool(streamId, WorkRPM.allowPipeFolding)
	
	streamWriteBool(streamId, spec.currentConfig)
	streamWriteBool(streamId, spec.allowTipping)
	streamWriteBool(streamId, spec.allowFolding)
	streamWriteBool(streamId, spec.allowRotating)
	streamWriteBool(streamId, spec.allowLifting)
	streamWriteBool(streamId, spec.allowMovingTools)
	streamWriteBool(streamId, spec.allowPipeFolding)
end

WorkRPMSetGlobalConfigEvent = {}
WorkRPMSetGlobalConfigEvent_mt = Class(WorkRPMSetGlobalConfigEvent, Event)
InitEventClass(WorkRPMSetGlobalConfigEvent, "WorkRPMSetGlobalConfigEvent")
function WorkRPMSetGlobalConfigEvent:emptyNew()
    local self = Event:new(WorkRPMSetGlobalConfigEvent_mt)
    self.className="WorkRPMSetGlobalConfigEvent"
    return self
end
function WorkRPMSetGlobalConfigEvent:new(object, allowTipping, allowFolding, allowRotating, allowLifting, allowMovingTools, allowPipeFolding)
	local self = WorkRPMSetGlobalConfigEvent:emptyNew()
	self.object = object
	self.allowTipping = allowTipping
	self.allowFolding = allowFolding
	self.allowRotating = allowRotating
	self.allowLifting = allowLifting
	self.allowMovingTools = allowMovingTools
	self.allowPipeFolding = allowPipeFolding
	return self
end
function WorkRPMSetGlobalConfigEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
    self.allowTipping = streamReadBool(streamId)
    self.allowFolding = streamReadBool(streamId)
    self.allowRotating = streamReadBool(streamId)
    self.allowLifting = streamReadBool(streamId)
    self.allowMovingTools = streamReadBool(streamId)
    self.allowPipeFolding = streamReadBool(streamId)
    self:run(connection)
end
function WorkRPMSetGlobalConfigEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
	streamWriteBool(streamId, self.allowTipping)
	streamWriteBool(streamId, self.allowFolding)
	streamWriteBool(streamId, self.allowRotating)
	streamWriteBool(streamId, self.allowLifting)
	streamWriteBool(streamId, self.allowMovingTools)
	streamWriteBool(streamId, self.allowPipeFolding)
end
function WorkRPMSetGlobalConfigEvent:run(connection)
	if self.object ~= nil then
		self.object:wrpmSetGlobalConfig(self.allowTipping, self.allowFolding, self.allowRotating, self.allowLifting, self.allowMovingTools, self.allowPipeFolding, true)
	end
	if not connection:getIsServer() then
		g_server:broadcastEvent(WorkRPMSetGlobalConfigEvent:new(self.object, self.allowTipping, self.allowFolding, self.allowRotating, self.allowLifting, self.allowMovingTools, self.allowPipeFolding), nil, connection, self.object)
	end
end
function WorkRPMSetGlobalConfigEvent.sendEvent(vehicle, allowTipping, allowFolding, allowRotating, allowLifting, allowMovingTools, allowPipeFolding, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(WorkRPMSetGlobalConfigEvent:new(vehicle, allowTipping, allowFolding, allowRotating, allowLifting, allowMovingTools, allowPipeFolding), nil, nil, vehicle)
		else
			g_client:getServerConnection():sendEvent(WorkRPMSetGlobalConfigEvent:new(vehicle, allowTipping, allowFolding, allowRotating, allowLifting, allowMovingTools, allowPipeFolding))
		end
	end
end

WorkRPMSetVehicleConfigEvent = {}
WorkRPMSetVehicleConfigEvent_mt = Class(WorkRPMSetVehicleConfigEvent, Event)
InitEventClass(WorkRPMSetVehicleConfigEvent, "WorkRPMSetVehicleConfigEvent")
function WorkRPMSetVehicleConfigEvent:emptyNew()
    local self = Event:new(WorkRPMSetVehicleConfigEvent_mt)
    self.className="WorkRPMSetVehicleConfigEvent"
    return self
end
function WorkRPMSetVehicleConfigEvent:new(object, currentConfig, allowTipping, allowFolding, allowRotating, allowLifting, allowMovingTools, allowPipeFolding)
	local self = WorkRPMSetVehicleConfigEvent:emptyNew()
	self.object = object
	self.currentConfig = currentConfig
	self.allowTipping = allowTipping
	self.allowFolding = allowFolding
	self.allowRotating = allowRotating
	self.allowLifting = allowLifting
	self.allowMovingTools = allowMovingTools
	self.allowPipeFolding = allowPipeFolding
	return self
end
function WorkRPMSetVehicleConfigEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
    self.currentConfig = streamReadBool(streamId)
    self.allowTipping = streamReadBool(streamId)
    self.allowFolding = streamReadBool(streamId)
    self.allowRotating = streamReadBool(streamId)
    self.allowLifting = streamReadBool(streamId)
    self.allowMovingTools = streamReadBool(streamId)
    self.allowPipeFolding = streamReadBool(streamId)
    self:run(connection)
end
function WorkRPMSetVehicleConfigEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
	streamWriteBool(streamId, self.currentConfig)
	streamWriteBool(streamId, self.allowTipping)
	streamWriteBool(streamId, self.allowFolding)
	streamWriteBool(streamId, self.allowRotating)
	streamWriteBool(streamId, self.allowLifting)
	streamWriteBool(streamId, self.allowMovingTools)
	streamWriteBool(streamId, self.allowPipeFolding)
end
function WorkRPMSetVehicleConfigEvent:run(connection)
	if self.object ~= nil then
		self.object:wrpmSetGlobalConfig(self.currentConfig, self.allowTipping, self.allowFolding, self.allowRotating, self.allowLifting, self.allowMovingTools, self.allowPipeFolding, true)
	end
	if not connection:getIsServer() then
		g_server:broadcastEvent(WorkRPMSetVehicleConfigEvent:new(self.object, self.currentConfig, self.allowTipping, self.allowFolding, self.allowRotating, self.allowLifting, self.allowMovingTools, self.allowPipeFolding), nil, connection, self.object)
	end
end
function WorkRPMSetVehicleConfigEvent.sendEvent(vehicle, currentConfig, allowTipping, allowFolding, allowRotating, allowLifting, allowMovingTools, allowPipeFolding, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(WorkRPMSetVehicleConfigEvent:new(vehicle, currentConfig, allowTipping, allowFolding, allowRotating, allowLifting, allowMovingTools, allowPipeFolding), nil, nil, vehicle)
		else
			g_client:getServerConnection():sendEvent(WorkRPMSetVehicleConfigEvent:new(vehicle, currentConfig, allowTipping, allowFolding, allowRotating, allowLifting, allowMovingTools, allowPipeFolding))
		end
	end
end

WorkRPMLowerImplementEvent = {}
WorkRPMLowerImplementEvent_mt = Class(WorkRPMLowerImplementEvent, Event)
InitEventClass(WorkRPMLowerImplementEvent, "WorkRPMLowerImplementEvent")
function WorkRPMLowerImplementEvent:emptyNew()
    local self = Event:new(WorkRPMLowerImplementEvent_mt)
    self.className="WorkRPMLowerImplementEvent"
    return self
end
function WorkRPMLowerImplementEvent:new(object)
	local self = WorkRPMLowerImplementEvent:emptyNew()
	self.object = object
	return self
end
function WorkRPMLowerImplementEvent:readStream(streamId, connection)
	self.object = NetworkUtil.readNodeObject(streamId)
    self:run(connection)
end
function WorkRPMLowerImplementEvent:writeStream(streamId, connection)
	NetworkUtil.writeNodeObject(streamId, self.object)
end
function WorkRPMLowerImplementEvent:run(connection)
	if self.object ~= nil then
		self.object.spec_workRPM.inputLowerImplement = true
	end
	if not connection:getIsServer() then
		g_server:broadcastEvent(WorkRPMLowerImplementEvent:new(self.object), nil, connection, self.object)
	end
end
function WorkRPMLowerImplementEvent.sendEvent(vehicle, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(WorkRPMLowerImplementEvent:new(vehicle), nil, nil, vehicle)
		else
			g_client:getServerConnection():sendEvent(WorkRPMLowerImplementEvent:new(vehicle))
		end
	end
end